# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: storage.proto

import sys

_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode("latin1"))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="storage.proto",
    package="storage_bench",
    syntax="proto3",
    serialized_options=None,
    serialized_pb=_b(
        '\n\rstorage.proto\x12\rstorage_bench"4\n\nObjectRead\x12\x12\n\nbucketName\x18\x01 \x01(\t\x12\x12\n\nobjectName\x18\x02 \x01(\t"J\n\x0bObjectWrite\x12\x12\n\nbucketName\x18\x01 \x01(\t\x12\x12\n\nobjectName\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65stination\x18\x03 \x01(\t"\x0f\n\rEmptyResponse2\x9d\x01\n\x13StorageBenchWrapper\x12\x43\n\x05Write\x12\x1a.storage_bench.ObjectWrite\x1a\x1c.storage_bench.EmptyResponse"\x00\x12\x41\n\x04Read\x12\x19.storage_bench.ObjectRead\x1a\x1c.storage_bench.EmptyResponse"\x00\x62\x06proto3'
    ),
)


_OBJECTREAD = _descriptor.Descriptor(
    name="ObjectRead",
    full_name="storage_bench.ObjectRead",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="bucketName",
            full_name="storage_bench.ObjectRead.bucketName",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="objectName",
            full_name="storage_bench.ObjectRead.objectName",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=32,
    serialized_end=84,
)


_OBJECTWRITE = _descriptor.Descriptor(
    name="ObjectWrite",
    full_name="storage_bench.ObjectWrite",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="bucketName",
            full_name="storage_bench.ObjectWrite.bucketName",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="objectName",
            full_name="storage_bench.ObjectWrite.objectName",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="destination",
            full_name="storage_bench.ObjectWrite.destination",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=86,
    serialized_end=160,
)


_EMPTYRESPONSE = _descriptor.Descriptor(
    name="EmptyResponse",
    full_name="storage_bench.EmptyResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=162,
    serialized_end=177,
)

DESCRIPTOR.message_types_by_name["ObjectRead"] = _OBJECTREAD
DESCRIPTOR.message_types_by_name["ObjectWrite"] = _OBJECTWRITE
DESCRIPTOR.message_types_by_name["EmptyResponse"] = _EMPTYRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ObjectRead = _reflection.GeneratedProtocolMessageType(
    "ObjectRead",
    (_message.Message,),
    {
        "DESCRIPTOR": _OBJECTREAD,
        "__module__": "storage_pb2"
        # @@protoc_insertion_point(class_scope:storage_bench.ObjectRead)
    },
)
_sym_db.RegisterMessage(ObjectRead)

ObjectWrite = _reflection.GeneratedProtocolMessageType(
    "ObjectWrite",
    (_message.Message,),
    {
        "DESCRIPTOR": _OBJECTWRITE,
        "__module__": "storage_pb2"
        # @@protoc_insertion_point(class_scope:storage_bench.ObjectWrite)
    },
)
_sym_db.RegisterMessage(ObjectWrite)

EmptyResponse = _reflection.GeneratedProtocolMessageType(
    "EmptyResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _EMPTYRESPONSE,
        "__module__": "storage_pb2"
        # @@protoc_insertion_point(class_scope:storage_bench.EmptyResponse)
    },
)
_sym_db.RegisterMessage(EmptyResponse)


_STORAGEBENCHWRAPPER = _descriptor.ServiceDescriptor(
    name="StorageBenchWrapper",
    full_name="storage_bench.StorageBenchWrapper",
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    serialized_start=180,
    serialized_end=337,
    methods=[
        _descriptor.MethodDescriptor(
            name="Write",
            full_name="storage_bench.StorageBenchWrapper.Write",
            index=0,
            containing_service=None,
            input_type=_OBJECTWRITE,
            output_type=_EMPTYRESPONSE,
            serialized_options=None,
        ),
        _descriptor.MethodDescriptor(
            name="Read",
            full_name="storage_bench.StorageBenchWrapper.Read",
            index=1,
            containing_service=None,
            input_type=_OBJECTREAD,
            output_type=_EMPTYRESPONSE,
            serialized_options=None,
        ),
    ],
)
_sym_db.RegisterServiceDescriptor(_STORAGEBENCHWRAPPER)

DESCRIPTOR.services_by_name["StorageBenchWrapper"] = _STORAGEBENCHWRAPPER

# @@protoc_insertion_point(module_scope)
